#include "gsl/gsl_rng.h"
#include "gsl/gsl_randist.h"


#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
double ran2(long *idum)
/*Long period (> 2 10^18) random number generator of L Ecuyer with Bays-Durham shuffle
and added safeguards. Returns a uniform random deviate between 0.0 and 1.0 (exclusive of
the endpoint values). Call with idum a negative integer to initialize; thereafter, do not alter
idum between successive deviates in a sequence. RNMX should approximate the largest ﬂoating
value that is less than 1.*/
{
	int j;
long k;
static long idum2=123456789;
static long iy=0;
static long iv[NTAB];
double temp;

if (*idum <= 0) {
if (-(*idum) < 1) *idum=1;
else *idum = -(*idum);
idum2=(*idum);
for (j=NTAB+7;j>=0;j--) {
k=(*idum)/IQ1;
*idum=IA1*(*idum-k*IQ1)-k*IR1;
if (*idum < 0) *idum += IM1;
if (j < NTAB) iv[j] = *idum;
}
iy=iv[0];
}
k=(*idum)/IQ1;
*idum=IA1*(*idum-k*IQ1)-k*IR1;
if (*idum < 0) *idum += IM1;
k=idum2/IQ2;
idum2=IA2*(idum2-k*IQ2)-k*IR2;
if (idum2 < 0) idum2 += IM2;
j=iy/NDIV;
iy=iv[j]-idum2;
iv[j] = *idum;
if (iy < 1) iy += IMM1;
if ((temp=(double)AM*iy) > RNMX) return (double) RNMX;
else return temp;
}


#undef IM1 
#undef IM2 
#undef AM 
#undef IMM1 
#undef IA1 
#undef IA2 
#undef IQ1 
#undef IQ2 
#undef IR1 
#undef IR2 
#undef NTAB 
#undef NDIV 
#undef EPS 
#undef RNMX 

double gammln(double xx)
{
	double x,tmp,ser;
	static double cof[6]={76.18009173,-86.50532033,24.01409822,
		-1.231739516,0.120858003e-2,-0.536382e-5};
	int j;
	x=xx-1.0;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.0;
	for (j=0;j<=5;j++) {
		x += 1.0;
		ser += cof[j]/x;
	}
	return -tmp+log(2.50662827465*ser);
}


#define ITMAX 100
#define EPS 3.0e-7
/*#define EPS 3.0e-7*/
void gcf(double *gammcf, double a, double x, double *gln)
{
	int n;
	double gold=0.0,g,fac=1.0,b1=1.0;
	double b0=0.0,anf,ana,an,a1,a0=1.0;
	double gammln();
	/*void mexErrMsgTxt();*/
	*gln=gammln(a);
	a1=x;
	for (n=1;n<=ITMAX;n++) {
		an=(double) n;
		ana=an-a;
		a0=(a1+a0*ana)*fac;
		b0=(b1+b0*ana)*fac;
		anf=an*fac;
		a1=x*a0+anf*a1;
		b1=x*b0+anf*b1;
		if (a1) {
			fac=1.0/a1;
			g=b1*fac;
			if (fabs((g-gold)/g) < EPS) {
				*gammcf=exp(-x+a*log(x)-(*gln))*g;
				return;
			}
			gold=g;
		}
	}
	mexErrMsgTxt("a too large, ITMAX too small in routine GCF");
}
#undef ITMAX
#undef EPS



#define ITMAX 100
#define EPS 3.0e-7
void gser(double *gamser, double a, double x, double *gln)
{
	int n;
	double sum,del,ap;
	double gammln();
	/*void mexErrMsgTxt();*/
	*gln=gammln(a);
	if (x <= 0.0) {
		if (x < 0.0) mexErrMsgTxt("x less than 0 in routine GSER");
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			ap += 1.0;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		mexErrMsgTxt("a too large, ITMAX too small in routine GSER");
		return;
	}
}
#undef ITMAX
#undef EPS


double gammp(double a, double x)
{
	double gamser,gammcf,gln;
	void gser(),gcf();
	if (x < 0.0 || a <= 0.0) mexErrMsgTxt("Invalid arguments in routine GAMMP");
	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return 1.0-gammcf;
	}
}



double norm(double x)
{
 double  gammp();
  if(x>=0.0) return 1.0+gammp(0.5, 0.5*x*x);
  else    return  1.0-1.0*gammp(0.5, 0.5*x*x);
}



double expdev(long *idum)
{
     double  dum;
     do
        dum=ran2(idum);
     while(dum==0.0);
     return -log(dum);
}



double gasdev(long  *idum)
{
	static int iset=0;
	static double gset;
	double fac,r,v1,v2;
	if  (iset == 0) {
		do {
			v1=2.0*ran2(idum)-1.0;
			v2=2.0*ran2(idum)-1.0;
			r=v1*v1+v2*v2;
		} while (r >= 1.0||r==0.0);
		fac=sqrt(-2.0*log(r)/r);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}



double onetn(double as,long *idum)
{
double aa, rz, zz, uu, vv;
/*long idum;*/
int flag=0;
if(as>=0.0)
   while(flag==0)
     {
     aa=0.5*as+0.5*sqrt(as*as+4.0);
     /*idum=rand();*/
     zz=expdev(idum)/aa+as;
     rz=-(zz-aa)*(zz-aa)/2.0;
     /*uu=rand()/(RAND_MAX+1.0);*/
     uu=ran2(idum);
     if(uu<=0.0) vv=uu;
     else  vv=log(uu);
     if(vv<=rz)  flag=1;
     }
     else
       while(flag==0)
          {
           /*idum=rand();*/
           zz=gasdev(idum);
           if(zz>as) flag=1;
          }
  return(zz);
}



double gamdev(int ia,long *idum)
{
	int j;
	double am,e,s,v1,v2,x,y;
	

	if (ia < 1) mexErrMsgTxt("Error in routine GAMDEV");
	if (ia < 6) {
		x=1.0;
		for (j=1;j<=ia;j++) x *= ran2(idum);
		x = -log(x);
	} else {
		do {
			do {
				do {
					v1=2.0*ran2(idum)-1.0;
					v2=2.0*ran2(idum)-1.0;
				} while (v1*v1+v2*v2 > 1.0);
				y=v2/v1;
				am=ia-1;
				s=sqrt(2.0*am+1.0);
				x=s*y+am;
			} while (x <= 0.0);
			e=(1.0+y*y)*exp(am*log(x/am)-s*y);
		} while (ran2(idum) > e);
	}
	return x;
}


double trunom(double as, double bl, long *idum)
{
  double  rz, zz, uu, vv, re=0.0;
  int flag=0;
  /*mexPrintf("as=%14f, bl=%14f\n", as, bl);*/
if(as>bl){
   mexPrintf("as=%14f, bl=%14f\n", as, bl);
   mexErrMsgTxt("Error: as>bl\n");  }
if(bl>=8)
   while(flag==0)
      {
       zz=onetn(as,idum);
       if(zz<= bl){
          flag=1;  re=zz;  }
      }
if(as<=-8)
   while(flag==0)
      {
       zz=-1.0*onetn(-1.0*bl,idum);
       if(zz>as){
          flag=1;  re=zz;  }
      }
if(as>-8 && bl<8.0)
  while(flag==0)
   {
   zz=as+(bl-as)*ran2(idum);
   if(bl<0.0)
     rz=(bl*bl-zz*zz)/2.0;
   else
     if(as>0.0)  rz=(as*as-zz*zz)/2.0;
       else   rz=-zz*zz/2.0;
   uu=ran2(idum);
   if(uu<=0.0)  vv=0.0;
     else   vv=log(uu);
   if(vv<=rz){
        flag=1;
        re=zz;
      }
   }
   return(re);
 }/* end */

 
 
 
 double onetn_gsl(double as, gsl_rng *r)
{
double aa, rz, zz, uu, vv;
/*long idum;*/
int flag=0;
if(as>=0.0)
   while(flag==0)
     {
     aa=0.5*as+0.5*sqrt(as*as+4.0);
     /*idum=rand();*/
     /*zz=expdev(idum)/aa+as;*/
     zz=gsl_ran_exponential (r, 1.0)/aa+as;     
     rz=-(zz-aa)*(zz-aa)/2.0;
     /*uu=rand()/(RAND_MAX+1.0);*/
     /*uu=ran2(idum);*/
     uu = gsl_rng_uniform(r);
     if(uu<=0.0) vv=uu;
     else  vv=log(uu);
     if(vv<=rz)  flag=1;
     }
     else
       while(flag==0)
          {
           /*idum=rand();*/
           /*zz=gasdev(idum);*/
           zz = gsl_ran_gaussian(r, 1.0);
           if(zz>as) flag=1;
          }
  return(zz);
}
 
 
 
 double trunom_gsl(double as, double bl, gsl_rng *r)
{
  double  rz, zz, uu, vv, re=0.0;
  int flag=0;
  /*mexPrintf("as=%14f, bl=%14f\n", as, bl);*/
if(as>bl){
   mexPrintf("as=%14f, bl=%14f\n", as, bl);
   mexErrMsgTxt("Error: as>bl\n");  }
if(bl>=8)
   while(flag==0)
      {
       /*zz=onetn(as,idum);*/
       zz=onetn_gsl(as,r);
       if(zz<= bl){
          flag=1;  re=zz;  }
      }
if(as<=-8)
   while(flag==0)
      {
       /*zz=-1.0*onetn(-1.0*bl,idum);*/
       zz=-1.0*onetn_gsl(-1.0*bl,r);
       if(zz>as){
          flag=1;  re=zz;  }
      }
if(as>-8 && bl<8.0)
  while(flag==0)
   {
   /*zz=as+(bl-as)*ran2(idum);*/
   zz=as+(bl-as)*gsl_rng_uniform(r);
   if(bl<0.0)
     rz=(bl*bl-zz*zz)/2.0;
   else
     if(as>0.0)  rz=(as*as-zz*zz)/2.0;
       else   rz=-zz*zz/2.0;
   /*uu=ran2(idum);*/
   uu = gsl_rng_uniform(r);
   if(uu<=0.0)  vv=0.0;
     else   vv=log(uu);
   if(vv<=rz){
        flag=1;
        re=zz;
      }
   }
   return(re);
 }/* end */

